program OSNP;

{ $DEFINE DEBUG}

uses
  Windows, Messages, SysUtils, Types, Math, Lite, Lite1, SmplCalcKrnl;

function IsCircleTrack(const Points: array of TPoint;
  var Center: TPoint): Boolean;
var
  I, Len: Integer;
  XX, YY: array of Double;
  Theta, ThetaLast, A, AA: Double;
begin
  Result := False;
  Len := Length(Points);
  if Len < 8 then
    Exit;
  SetLength(XX, Len);
  SetLength(YY, Len);
  for I := 0 to Len - 1 do
    with Points[I] do
    begin
      XX[I] := X;
      YY[I] := Y;
    end;
  Center.X := Round((MinValue(XX) + MaxValue(XX)) / 2);
  Center.Y := Round((MinValue(YY) + MaxValue(YY)) / 2);
  AA := 0;
  ThetaLast := 0;
  for I := 0 to Len - 1 do
  begin
    Theta := ArcTan2(XX[I] - Center.X, Center.Y - YY[I]);
    if I > 0 then
    begin
      A := Theta - ThetaLast;
      if InRange(Abs(A), Pi, 2 * Pi) then
        if A > 0 then
          A := A - Pi * 2
        else
          A := A + Pi * 2;
      if Abs(A) > (Pi / 2) then
        Exit;
      AA := AA + A;
    end;
    ThetaLast := Theta;
  end;
  Result := Abs(AA) > 2 * Pi;      
end;

procedure PasteSelect(const S: string);
var
  I: Integer;
begin
  SetClipboard(CF_TEXT, PChar(S)^, Length(S) + 1);
  keybd_event(VK_CONTROL, 0, 0, 0);
  SimulateKeystroke(Ord('V'));
  keybd_event(VK_CONTROL, 0, KEYEVENTF_KEYUP, 0);
  keybd_event(VK_SHIFT, 0, 0, 0);
  for I := 1 to Length(S) do
    SimulateKeystroke(VK_LEFT);
  keybd_event(VK_SHIFT, 0, KEYEVENTF_KEYUP, 0);
end;

function SelectCopyToken: string;
var
  S: string;
  I, J: Integer;
begin
  keybd_event(VK_CONTROL, 0, 0, 0);
  SimulateKeystroke(VK_RIGHT);
  keybd_event(VK_SHIFT, 0, 0, 0);
  SimulateKeystroke(VK_LEFT);
  keybd_event(VK_SHIFT, 0, KEYEVENTF_KEYUP, 0);
  SimulateKeystroke(Ord('C'));
  keybd_event(VK_CONTROL, 0, KEYEVENTF_KEYUP, 0);
  S := TextInClipboard;
  I := Length(S);
  while (I > 0) and (S[I] <= ' ') do
    Dec(I);
  for J := 1 to I do
    SimulateKeystroke(VK_RIGHT);
  keybd_event(VK_CONTROL, 0, 0, 0);
  keybd_event(VK_SHIFT, 0, 0, 0);
  SimulateKeystroke(VK_LEFT);
  keybd_event(VK_SHIFT, 0, KEYEVENTF_KEYUP, 0);
  SimulateKeystroke(Ord('C'));
  keybd_event(VK_CONTROL, 0, KEYEVENTF_KEYUP, 0);
  Result := TextInClipboard;
end;

type
  TCalcOp = (coNone, coResult, coAdd, coSubtract, coMutliply, coDivide);
  TButtonRec = record
    Cap: WideString;
    Num: Integer;
    Op: TCalcOp;
  end;

var
  ButtonData: array[1..4, 0..3] of TButtonRec = (
    ((Cap: '7'; Num: 7), (Cap: '8'; Num: 8), (Cap: '9'; Num: 9),
      (Cap: WideChar($F7); Op: coDivide)),
    ((Cap: '4'; Num: 4), (Cap: '5'; Num: 5), (Cap: '6'; Num: 6),
      (Cap: WideChar($D7); Op: coMutliply)),
    ((Cap: '1'; Num: 1), (Cap: '2'; Num: 2), (Cap: '3'; Num: 3),
      (Cap: '-'; Op: coSubtract)),
    ((Cap: '0'; Num: 0), (Num: -1), (Cap: '='; Op: coResult),
      (Cap: '+'; Op: coAdd))
  );
  BtnSize: TSize;
  CalcMode: Boolean;

procedure RequiredValidBtnSize;
begin
  with BtnSize do
  begin
    cy := GetSystemMetrics(SM_CYMENU);
    cx := MulDiv(cy, 4, 3);
  end;
end;

function HighBtnPosX: Integer;
begin
  Result := 2 + Ord(CalcMode);
end;

type
  TFontParams = record
    Name: string;
    Size: Integer;
    Bold, Italic: Boolean;
  end;

const
  DefFontParams: TFontParams = (Name: 'Verdana'; Size: 8; Bold: True);

var
  Margin, BtnMargin, CalcEditBottomMargin: Integer;
  CornerCurve, BtnCornerCurve, CalcEditCornerCurve: TSize;
  BGColor, PenColor, BtnColor, BtnTextColor, BtnHoverColor, BtnHoverTextColor,
    CalcEditColor, CalcEditTextColor, MaskColor: TColorRef;
  FontParams: TFontParams;
  UseHandCursor, ClickOnDown: Boolean;
  Font: HFONT;

procedure InitParams;
var
  OldFontParams: TFontParams;
  lf: TLogFont;
begin
  OldFontParams := FontParams;
  Margin := 5;
  CornerCurve.cx := 5;
  CornerCurve.cy := 5;
  BGColor := 0;
  PenColor := $C0C0C0;
  FontParams := DefFontParams;
  BtnMargin := 1;
  BtnCornerCurve.cx := 5;
  BtnCornerCurve.cy := 5;
  BtnColor := 0;
  BtnTextColor := $00FFFF;
  BtnHoverColor := 0;
  BtnHoverTextColor := $FFFFFF;
  UseHandCursor := False;
  ClickOnDown := True;
  CalcEditBottomMargin := 3;
  CalcEditCornerCurve.cx := 5;
  CalcEditCornerCurve.cy := 5;
  CalcEditColor := 0;
  CalcEditTextColor := $FFFFFF;
  MaskColor := $FF00FF;
  try
    ReadIniFile([
      MakeParamRec('Margin', @Margin, dtInteger),
      MakeParamRec('CornerCurveX', @CornerCurve.cx, dtInteger),
      MakeParamRec('CornerCurveY', @CornerCurve.cy, dtInteger),
      MakeParamRec('BGColor', @BGColor, dtInteger),
      MakeParamRec('PenColor', @PenColor, dtInteger),
      MakeParamRec('FontFace', @FontParams.Name, dtString),
      MakeParamRec('FontSize', @FontParams.Size, dtInteger),
      MakeParamRec('BoldFont', @FontParams.Bold, dtBoolean),
      MakeParamRec('ItalicFont', @FontParams.Italic, dtBoolean),
      MakeParamRec('BtnMargin', @BtnMargin, dtInteger),
      MakeParamRec('BtnCornerCurveX', @BtnCornerCurve.cx, dtInteger),
      MakeParamRec('BtnCornerCurveY', @BtnCornerCurve.cy, dtInteger),
      MakeParamRec('BtnColor', @BtnColor, dtInteger),
      MakeParamRec('BtnTextColor', @BtnTextColor, dtInteger),
      MakeParamRec('BtnHoverColor', @BtnHoverColor, dtInteger),
      MakeParamRec('BtnHoverTextColor', @BtnHoverTextColor, dtInteger),
      MakeParamRec('UseHandCursor', @UseHandCursor, dtBoolean),
      MakeParamRec('ClickOnDown', @ClickOnDown, dtBoolean),
      MakeParamRec('CalcEditBottomMargin', @CalcEditBottomMargin, dtInteger),
      MakeParamRec('CalcEditCornerCurveX', @CalcEditCornerCurve.cx, dtInteger),
      MakeParamRec('CalcEditCornerCurveY', @CalcEditCornerCurve.cy, dtInteger),
      MakeParamRec('CalcEditColor', @CalcEditColor, dtInteger),
      MakeParamRec('CalcEditTextColor', @CalcEditTextColor, dtInteger),
      MakeParamRec('MaskColor', @MaskColor, dtInteger)
    ]);
  except
    on EInOutError do;
  end;
  if not CompareMem(@FontParams, @OldFontParams, SizeOf(TFontParams)) then
  begin
    if Font <> 0 then
      DeleteObject(Font);
    FillChar(lf, SizeOf(TLogFont), 0);
    with lf, FontParams do
    begin
      lfHeight := -MulDiv(Size, GetDeviceCaps(GetDC(0), LOGPIXELSY), 72);
      if Bold then
        lfWeight := FW_BOLD
      else
        lfWeight := FW_NORMAL;
      if Italic then
        lfItalic := Ord(True);
      StrPLCopy(lfFaceName, Name, LF_FACESIZE - 1);
    end;
    Font := CreateFontIndirect(lf);
  end;
end;

function ValidWndSize: TSize;
begin
  Result.cx := BtnSize.cx * (HighBtnPosX + 1) + Margin * 2;
  Result.cy := BtnSize.cy * (High(ButtonData) + 1) + Margin * 2;
end;

function BtnRect(PosX, PosY: Integer): TRect;
begin
  with Result, BtnSize do
  begin
    Left := PosX * cx;
    Top := PosY * cy;
    Right := (PosX + 1) * cx;
    Bottom := (PosY + 1) * cy;
  end;
  OffsetRect(Result, Margin, Margin);
  InflateRect(Result, -BtnMargin, -BtnMargin);
end;

function CalcEditRect: TRect;
begin
  with Result, BtnSize do
  begin
    Left := 0;
    Top := 0;
    Right := cx * 4;
    Bottom := cy - CalcEditBottomMargin;
  end;
  OffsetRect(Result, Margin, Margin);
end;

function CalcEditTextRect: TRect;
begin
  Result := CalcEditRect;
  InflateRect(Result, -Max(MulDiv(CalcEditCornerCurve.cx, 2, 5), 3), -1);
end;

procedure RoundRect1(DC: HDC; const Rect: TRect; const CornerCurve: TSize;
  Color: TColorRef);
begin
  SetDCBrushColor(DC, Color);
  with Rect, CornerCurve do
    RoundRect(DC, Left, Top, Right, Bottom, cx, cy);
end;

const
  M_CALCCHANGE = WM_USER;

var
  MouseTrackBuf: array[Byte] of TPoint;
  MouseTrackCount, MouseIdleCount: Integer;
  CurBtnPosX, CurBtnPosY, OldHoverPosX, OldHoverPosY: Integer;
  MouseOnButton, OldMouseOnButton, Clicked, OldClicked, Tracking: Boolean;
  tme: TTrackMouseEvent = (cbSize: SizeOf(TTrackMouseEvent);
    dwFlags: TME_LEAVE);
  ClickPoint: TSmallPoint;
  Calc: TSmplCalcKrnl;
  CalcEdit: string;
  SuppressCalcEditUpdate: Boolean;

function WindowProc(Wnd: HWND; Msg, WParam: Cardinal; LParam: Longint): Longint;
  stdcall;
var
  DC: HDC;
  DCInitialized: Boolean;

  procedure InitializeDC(DC: HDC);
  begin
    SelectObject(DC, GetStockObject(DC_BRUSH));
    SelectObject(DC, Font);
    SelectObject(DC, GetStockObject(DC_PEN));
    SetDCPenColor(DC, PenColor);
  end;

  procedure RequiredValidDC;
  begin
    if DC = 0 then
      DC := GetDC(Wnd);
    if not DCInitialized then
    begin
      InitializeDC(DC);
      DCInitialized := True;
    end;
  end;

  procedure DrawControl(const Rect: TRect; const CornerCurve: TSize;
    const TextRect: TRect; Color, TextColor: TColorRef;
    const Caption: WideString; Alignment: Cardinal);
  var
    rc: TRect;
  begin
    RequiredValidDC;
    RoundRect1(DC, Rect, CornerCurve, Color);
    SetTextColor(DC, TextColor);
    SetBkColor(DC, Color);
    rc := TextRect;
    DrawTextW(DC, PWideChar(Caption), Length(Caption), rc, Alignment +
      DT_VCENTER + DT_SINGLELINE);
  end;

  procedure DrawButton(PosX, PosY: Integer);
  var
    rcText: TRect;
    Pointed: Boolean;
    Colors: array[Boolean, Boolean] of TColorRef;
    TextColors: array[Boolean, Boolean] of TColorRef;
  begin
{$IFDEF DEBUG}
   if(PosX<0)or(PosX>3)or(PosY<1)or(PosY>4)then begin MessageBeep(0);Exit end;
{$ENDIF}
    rcText := BtnRect(PosX, PosY);
    Colors[False, False] := BtnColor;
    Colors[False, True] := BtnColor;
    Colors[True, False] := BtnHoverColor;
    Colors[True, True] := Hue(Random * 360);
    TextColors[False, False] := BtnTextColor;
    TextColors[False, True] := BtnTextColor;
    TextColors[True, False] := BtnHoverTextColor;
    TextColors[True, True] := $000000;
    Pointed := MouseOnButton and (PosX = CurBtnPosX) and (PosY = CurBtnPosY);
    DrawControl(BtnRect(PosX, PosY), BtnCornerCurve, rcText, Colors[Pointed,
      Clicked], TextColors[Pointed, Clicked], ButtonData[PosY, PosX].Cap,
      DT_CENTER);
  end;

  procedure DrawCalcEdit;
  begin
    DrawControl(CalcEditRect, CalcEditCornerCurve, CalcEditTextRect,
      CalcEditColor, CalcEditTextColor, CalcEdit, DT_RIGHT);
  end;

const
  Cursors: array[Boolean] of MakeIntResource = (IDC_ARROW, IDC_HAND);

  procedure MouseAct;
  begin
    if (CurBtnPosX <> OldHoverPosX) or (CurBtnPosY <> OldHoverPosY) or
      (MouseOnButton <> OldMouseOnButton) or (Clicked <> OldClicked) then
    begin
      if UseHandCursor and (MouseOnButton <> OldMouseOnButton) then
        SetCursor(LoadCursor(0, Cursors[MouseOnButton]));
      if OldMouseOnButton then
        DrawButton(OldHoverPosX, OldHoverPosY);
      if MouseOnButton then
        DrawButton(CurBtnPosX, CurBtnPosY);
      OldHoverPosX := CurBtnPosX;
      OldHoverPosY := CurBtnPosY;
      OldMouseOnButton := MouseOnButton;
      OldClicked := Clicked;
    end;
  end;

  procedure UpdateCalcEdit;
  var
    S: string;
    Size: TSize;
    Prec: Integer;
  begin
    RequiredValidDC;
    Prec := 19;
    with CalcEditTextRect do
      repeat
        S := Format('%.*g', [Prec, Calc.EditValue]);
        GetTextExtentPoint32(DC, PChar(S), Length(S), Size);
        Dec(Prec);
      until Size.cx <= Right - Left;
    CalcEdit := S;
    DrawCalcEdit;
  end;

  procedure UpdWndSize;
  var
    rc: TRect;
  begin
    GetClientRect(Wnd, rc);
    with ValidWndSize do
      if (cx <> rc.Right) or (cy <> rc.Left) then
        SetWindowPos(Wnd, HWND_TOP, 0, 0, cx, cy, SWP_NOMOVE or SWP_NOACTIVATE);
  end;

  procedure BtnClick(const ButtonRec: TButtonRec);

    function EditHaveRoomForStr(const S: string): Boolean;
    var
      Size: TSize;
    begin
      RequiredValidDC;
      GetTextExtentPoint32(DC, PChar(S), Length(S), Size);
      with CalcEditTextRect do
        Result := Size.cx <= Right - Left;
    end;

  var
    Ch: Char;
    S: string;
  begin
    with ButtonRec do
      if CalcMode then
        case Op of
          coNone:
            begin
              S := CalcEdit;
              if not (csValueEdited in Calc.State) then
                S := '';
              Ch := #0;
              if Num >= 0 then
                Ch := Chr(Ord('0') + Num)
              else
                if Pos(DecimalSeparator, S) = 0 then
                  Ch := DecimalSeparator;
              if (Ch <> #0) and EditHaveRoomForStr(S + Ch) then
              begin
                S := S + Ch;
                SuppressCalcEditUpdate := True;
                try
                  Calc.EditValue := StrToFloat(S);
                finally
                  SuppressCalcEditUpdate := False;
                end;
                CalcEdit := S;
                DrawCalcEdit;
              end
              else
                MessageBeep(0);
            end;
          coAdd..coDivide: Calc.Op(TSmplCalcOp(Ord(Op) - Ord(coAdd)));
          coResult:
            begin
              Calc.Evaluate;
              PasteSelect(FloatToStr(Calc.Evaluation));
            end;
        end
      else
        case Op of
          coNone:
            begin
              if Num >= 0 then
                SimulateKeystroke(VK_NUMPAD0 + Num)
              else
                SimulateKeystroke(VK_DECIMAL);
            end;
          coResult:
            begin
              CalcMode := True;
              Calc.Value := StrToFloatDef(SelectCopyToken, 0);
              UpdWndSize;
            end;
        end;
  end;

  procedure Click;
  begin
    if MouseOnButton then
      BtnClick(ButtonData[CurBtnPosY, CurBtnPosX]);
  end;

  procedure Show(const Point: TPoint);
  var
    Cursor: HCURSOR;
  begin
    InitParams;
    if UseHandCursor then
      Cursor := 0
    else
      Cursor := LoadCursor(0, IDC_ARROW);
    if GetClassLong(Wnd, GCL_HCURSOR) <> Cursor then
      SetClassLong(Wnd, GCL_HCURSOR, Cursor);
    RequiredValidBtnSize;
    CalcMode := False;
    with ValidWndSize do
      SetWindowPos(Wnd, HWND_TOP, Point.X - MulDiv(BtnSize.cx, 3, 2) -
        Margin, Point.Y - (cy - MulDiv(BtnSize.cy, 5, 2)) + Margin, cx, cy,
        SWP_NOACTIVATE);
    ShowWindow(Wnd, SW_SHOWNA);
  end;
  
  procedure ShowHide(const Point: TPoint);
  begin
    if IsWindowVisible(Wnd) then
      ShowWindow(Wnd, SW_HIDE)
    else
      Show(Point);
  end;

var
  ps: TPaintStruct;
  PosX, PosY: Integer;
  Dummy, rc: TRect;
  MemDC: HDC;
  MemBM: HBITMAP;
  pt: Tpoint;
  MousePos: TSmallPoint absolute LParam;
begin
  Result := 0;
  DC := 0;
  DCInitialized := False;
  try
    if (Msg >= WM_MOUSEFIRST) and (Msg <= WM_MOUSELAST) then
    begin
      if Msg = WM_LBUTTONUP then
      begin
        if not ClickOnDown then
          Click;
        Clicked := False;
        ReleaseCapture;
      end;
      if not Clicked then
        with MousePos do
        begin
          CurBtnPosX := Floor((X - Margin) / BtnSize.cx);
          CurBtnPosY := Floor((Y - Margin) / BtnSize.cy);
          MouseOnButton := (CurBtnPosY >= Low(ButtonData)) and (CurBtnPosY <=
            High(ButtonData)) and (CurBtnPosX >= Low(ButtonData[CurBtnPosY]))
            and (CurBtnPosX <= HighBtnPosX);
        end;
      if Msg = WM_LBUTTONDOWN then
      begin
        SetCapture(Wnd);
        Clicked := True;
        ClickPoint := MousePos;
        if ClickOnDown then
          Click;
      end;
      MouseAct;
    end;
    case Msg of
{$IFDEF DEBUG}
      WM_CREATE: Show(Point(800, 600));
{$ENDIF}
      WM_PAINT:
        begin
          RequiredValidBtnSize;
          UpdWndSize;
          DC := BeginPaint(Wnd, ps);
          try
            for PosY := Low(ButtonData) to High(ButtonData) do
              for PosX := Low(ButtonData[PosY]) to HighBtnPosX do
                if IntersectRect(Dummy, BtnRect(PosX, PosY), ps.rcPaint) then
                  DrawButton(PosX, PosY);
            if CalcMode and IntersectRect(Dummy, CalcEditRect, ps.rcPaint) then
              DrawCalcEdit;
          finally
            DC := 0;
            EndPaint(Wnd, ps);
          end;
        end;
      WM_ERASEBKGND:
        begin
          GetClientRect(Wnd, rc);
          MemDC := CreateCompatibleDC(WParam);
          MemBM := CreateCompatibleBitmap(WParam, rc.Right, rc.Bottom);
          SelectObject(MemDC, MemBM);
          try
            InitializeDC(MemDC);
            SetDCBrushColor(MemDC, MaskColor);
            FillRect(MemDC, rc, GetStockObject(DC_BRUSH));
            if not CalcMode then
              rc.Top := BtnSize.cy;
            RoundRect1(MemDC, rc, CornerCurve, BGColor);
            BitBlt(WParam, 0, 0, rc.Right, rc.Bottom, MemDC, 0, 0, SRCCOPY);
          finally
            DeleteDC(MemDC);
            DeleteObject(MemBM);
          end;
          Result := 1;
        end;
      WM_TIMER:
        begin
          GetCursorPos(pt);
          if ((pt.X <> MouseTrackBuf[0].X) or (pt.Y <> MouseTrackBuf[0].Y)) then
          begin
            MouseTrackCount := Min(MouseTrackCount, Length(MouseTrackBuf) - 1);
            Move(MouseTrackBuf[0], MouseTrackBuf[1], MouseTrackCount *
              SizeOf(TPoint));
            MouseTrackBuf[0] := pt;
            Inc(MouseTrackCount);
            if IsCircleTrack(Slice(MouseTrackBuf, MouseTrackCount), pt) then
            begin
              ShowHide(pt);
              MouseTrackCount := 0;
            end;
            MouseIdleCount := 0;
          end
          else
          begin
            Inc(MouseIdleCount);
            if MouseIdleCount >= 3 then
              MouseTrackCount := 0;
          end;
        end;
      WM_MOUSEMOVE:
        begin
          if Clicked and not MouseOnButton then
          begin
            GetCursorPos(pt);
            SetWindowPos(Wnd, HWND_TOP, pt.X - ClickPoint.x, pt.Y -
              ClickPoint.y, 0, 0, SWP_NOSIZE or SWP_NOACTIVATE);
          end;  
          if not Tracking then
            Tracking := TrackMouseEvent(tme);
        end;
      WM_RBUTTONDOWN: ShowWindow(Wnd, SW_HIDE);
      WM_MBUTTONDOWN:
        begin
          CalcMode := not CalcMode;
          if CalcMode then
            Calc.Value := 0
          else
            if OldHoverPosX > 2 then
              OldMouseOnButton := False;
          UpdWndSize;
        end;
      WM_MOUSELEAVE:
        begin
          MouseOnButton := False;
          MouseAct;
          Tracking := False;
        end;
      M_CALCCHANGE:
        if not SuppressCalcEditUpdate then
          UpdateCalcEdit;
    else
      Result := DefWindowProc(Wnd, Msg, WParam, LParam);
    end;
  finally
    if DC <> 0 then
      ReleaseDC(Wnd, DC);
  end;
end;

var
  Cls: WNDCLASS = (lpfnWndProc: @WindowProc; lpszClassName: 'OSNPWnd');
  Wnd: HWND;
  Msg: TMsg;

type
  TFakeClass = class
    procedure CalcChange(Sender: TObject);
  end;

procedure TFakeClass.CalcChange(Sender: TObject);
begin
  SendMessage(HWND(Self), M_CALCCHANGE, 0, 0);
end;

begin
  if WaitForSingleObject(CreateMutex(nil, True, 'OSNP_Running'), 0) =
    WAIT_OBJECT_0 then
  begin
    InitParams;
    ButtonData[4, 1].Cap := DecimalSeparator;
    Calc := TSmplCalcKrnl.Create;
    Cls.hInstance := HInstance;
    if not UseHandCursor then
      Cls.hCursor := LoadCursor(0, IDC_ARROW);
    Cls.hbrBackground := GetStockObject(DC_BRUSH);
    Windows.RegisterClass(Cls);
    Wnd := CreateWindowEx(WS_EX_TOPMOST or WS_EX_LAYERED or WS_EX_NOACTIVATE,
      Cls.lpszClassName, '', WS_POPUP, 0, 0, 0, 0, 0, 0, HInstance, nil);
    SetLayeredWindowAttributes(Wnd, MaskColor, 128, LWA_COLORKEY or LWA_ALPHA);
    tme.hwndTrack := Wnd;
    Calc.OnChange := TFakeClass(Wnd).CalcChange;
    SetTimer(Wnd, 0, 10, nil);
    Randomize;
    while GetMessage(Msg, 0, 0, 0) do
      DispatchMessage(Msg);
  end;
end.
