unit SmplCalcKrnl;

interface

type
  TSmplCalcOp = (scoAdd, scoSubtract, scoMutliply, scoDivide);
  TSmplCalcState = set of (csValueEdited, csEvaluated);

  TCalcOpCacheRec = record
    Valid, Neg: Boolean;
    Value: Extended;
  end;

  TNotifyEvent = procedure(Sender: TObject) of object;

  TSmplCalcKrnl = class
  private
    FValue: Extended;
    FEvaluation: Extended;
    FState: TSmplCalcState;
    FOpCache: array[0..1] of TCalcOpCacheRec;
    FOpLevel: Integer;
    FOnChange: TNotifyEvent;
    function GetEditValue: Extended;
    procedure SetEditValue(const Value: Extended);
    procedure SetValue(const Value: Extended);
  protected
    procedure Changed; virtual;
    procedure ClearCache(Level: Integer = 0);
  public
    procedure Evaluate(Level: Integer = 0);
    procedure Op(AOp: TSmplCalcOp);
    property EditValue: Extended read GetEditValue write SetEditValue;
    property Evaluation: Extended read FEvaluation;
    property State: TSmplCalcState read FState;
    property Value: Extended read FValue write SetValue;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

implementation

{ TSmplCalcKrnl }

procedure TSmplCalcKrnl.Changed;
begin
  if Assigned(FOnChange) then
    FOnChange(Self);
end;

procedure TSmplCalcKrnl.ClearCache(Level: Integer);
var
  I: Integer;
begin
  for I := High(FOpCache) downto Level do
    FOpCache[I].Valid := False;
  FOpLevel := Level;
end;

procedure TSmplCalcKrnl.Evaluate(Level: Integer);
var
  I: Integer;
  Saved8087CW: Word;
begin
  FEvaluation := FValue;
  for I := High(FOpCache) downto Level do
    with FOpCache[I] do
      if Valid then
      begin
        FValue := FEvaluation;
        Saved8087CW := Default8087CW;
        try
          Set8087CW($133f); { Disable all fpu exceptions }
          case I * 2 + Ord(Neg) of
            0: FEvaluation := Value + FValue;
            1: FEvaluation := Value - FValue;
            2: FEvaluation := Value * FValue;
            3: FEvaluation := Value / FValue;
          end;
        finally
          Set8087CW(Saved8087CW);
        end;
        Value := FEvaluation;
        if I < High(FOpCache) then
          FOpCache[I + 1].Valid := False;
      end;
  FState := [csEvaluated];
  Changed;
end;

function TSmplCalcKrnl.GetEditValue: Extended;
begin
  if csEvaluated in FState then
    Result := FEvaluation
  else
    Result := FValue;
end;

procedure TSmplCalcKrnl.Op(AOp: TSmplCalcOp);
var
  Level: Integer;
begin
  if csEvaluated in FState then
  begin
    ClearCache;
    FValue := FEvaluation;
  end else
  if not (csValueEdited in FState) then
    ClearCache(FOpLevel);
  Level := Ord(AOp) div 2;
  Evaluate(Level);
  ClearCache(Level);
  with FOpCache[Level] do
  begin
    Value := FEvaluation;
    Neg := Boolean(Ord(AOp) and 1);
    Valid := True;
  end;
  FValue := FEvaluation;
  FState := [];
end;

procedure TSmplCalcKrnl.SetEditValue(const Value: Extended);
begin
  if csEvaluated in FState then
    ClearCache;
  FValue := Value;
  FState := [csValueEdited];
  Changed;
end;

procedure TSmplCalcKrnl.SetValue(const Value: Extended);
begin
  ClearCache;
  FValue := Value;
  FState := [];
  Changed;
end;

end.
